unit unextfloattestmain;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, StdCtrls, ComCtrls,
  ExtCtrls, Math, unextfloat, umsvcrtdll;

type

  { TForm1 }

  TForm1 = class(TForm)
    btnStart: TButton;
    Label1: TLabel;
    lblResult: TLabel;
    Memo1: TMemo;
    Panel1: TPanel;
    pbProgress: TProgressBar;
    procedure btnStartClick(Sender: TObject);
  private
    function TestFromNegToPos(const AStringList : TStrings) : Boolean;
    function TestFromPosToNeg(const AStringList : TStrings) : Boolean;

  public

  end;

var
  Form1: TForm1;

implementation

{$R *.lfm}

{ TForm1 }

procedure TForm1.btnStartClick(Sender: TObject);
var
  sl : TStringList;
begin
  lblResult.Caption := '1 of 2: Running from negative to positive';
  lblResult.Repaint;
  sl := TStringList.Create;
  try
    sl.Add('Begin of test');
    sl.Add('Test 1 of 2: Running from -Infinity to MaxDouble');

    if TestFromNegToPos(sl) then
    begin
      sl.Add('Test 1 of 1: Finished');
      sl.Add('Test 1 of 2: Running from Infinity to -MaxDouble');
      lblResult.Caption := '2 of 2: Running from positive to negative';
      if TestFromPosToNeg(sl) then
      begin
        lblResult.Caption := 'Success';
        sl.Add('Test 2 of 2: Finished');
      end
      else
        lblResult.Caption := 'Failed';
    end
    else
      lblResult.Caption := 'Failed';

    Memo1.Lines.BeginUpdate;
    try
      sl.Add('End of tests');
      Memo1.Lines.Assign(sl);
    finally
      Memo1.Lines.EndUpdate;
    end;
    pbProgress.Position := pbProgress.Max;
  finally
    sl.Free;
  end;
end;

function TForm1.TestFromNegToPos(const AStringList: TStrings): Boolean;
const
  ExpectedTests = 4283;
  loops = 1000;
var
  ddll, dfpc : Double;
  ds, dd : Double;
  i : Integer;
  cnt : Integer;
  d : Double;
  s : String;

begin
  Result := False;
  cnt := 0;
  dd := MaxDouble;
  ds := -Infinity;
  repeat
    d := cnt / ExpectedTests;
    if d > 1.0 then
      d := 1.0;
    pbProgress.Position := Trunc(d* 100.0);
    if (Abs(ds) > 0.5) and (Abs(ds) < 1e6) then
      s := Format('[%5.5d] Start %1.30f',[cnt, ds])
    else
      s := Format('[%5.5d] Start %1.30g',[cnt, ds]);


    for i := 0 to loops-1 do
    begin
      if ds = 0.0 then
        s := s + ' (Testing 0.0) ';
      ddll := umsvcrtdll.nextafter(ds,dd);
      dfpc := unextfloat.NextAfter(ds,dd);
      if ddll <> dfpc then
      begin
        if (Abs(ds) > 0.5) and (Abs(ds) < 1e6) then
          s := s + Format(' Stop %1.30f',[ds])
        else
          s := s + Format(' Stop %1.30g',[ds]);
        AStringList.Add(s);
        AStringList.Add('Test Failed!');
        Exit;
      end;
      if dfpc = MaxDouble then Break;
      ds := dfpc;
    end;
    if dfpc = MaxDouble then
    begin
{
      // This failes with umsvcrtdll.nextafter, causing a floating point violation
      ddll := umsvcrtdll.nextafter(MaxDouble,Infinity);
      // This succeeds with returning infinity
      dfpc := unextfloat.NextAfter(MaxDouble,Infinity);
}
      s := s + Format(' Stop %1.30g (MaxDouble)',[ds]);
      AStringList.Add(s);
      AStringList.Add('Test succeeded');
      Result := True;
      Break;
    end;
    if (Abs(ds) > 0.5) and (Abs(ds) < 1e6) then
      s := s + Format(' Stop %1.30f',[ds])
    else
      s := s + Format(' Stop %1.30g',[ds]);
    AStringList.Add(s);

    if ds > 0 then
    begin
      if ds > MaxDouble / 2.0 then
        ds := ds + ((MaxDouble - ds) / 2.0)
      else
        ds := ds * 2.0;
    end
    else
    begin
      if ds < -MaxDouble / 2.0 then
        ds := ds + ((MaxDouble - Abs(ds)) / 2.0)
      else
        ds := ds / 2.0;
    end;
    Inc(cnt)
  until False;
end;

function TForm1.TestFromPosToNeg(const AStringList: TStrings): Boolean;
const
  ExpectedTests = 4283;
  loops = 1000;
var
  ddll, dfpc : Double;
  ds, dd : Double;
  i : Integer;
  cnt : Integer;
  d : Double;
  s : String;

begin
  Result := False;
  cnt := 0;
  dd := -MaxDouble;
  ds := Infinity;
  repeat
    d := cnt / ExpectedTests;
    if d > 1.0 then
      d := 1.0;
    pbProgress.Position := Trunc(d* 100.0);
    if (Abs(ds) > 0.5) and (Abs(ds) < 1e6) then
      s := Format('[%5.5d] Start %1.30f',[cnt, ds])
    else
      s := Format('[%5.5d] Start %1.30g',[cnt, ds]);


    for i := 0 to loops-1 do
    begin
      if ds = 0.0 then
        s := s + ' (Testing 0.0) ';
      ddll := umsvcrtdll.nextafter(ds,dd);
      dfpc := unextfloat.NextAfter(ds,dd);
      if ddll <> dfpc then
      begin
        if (Abs(ds) > 0.5) and (Abs(ds) < 1e6) then
          s := s + Format(' Stop %1.30f',[ds])
        else
          s := s + Format(' Stop %1.30g',[ds]);
        AStringList.Add(s);
        AStringList.Add('Test Failed!');
        Exit;
      end;
      if dfpc = -MaxDouble then Break;
      ds := dfpc;
    end;
    if dfpc = -MaxDouble then
    begin

{
      // This failes with umsvcrtdll.nextafter, causing a floating point violation
      ddll := umsvcrtdll.nextafter(-MaxDouble,-Infinity);
      // This succeeds with returning -Infinity
      dfpc := unextfloat.NextAfter(-MaxDouble,-Infinity);
}
      s := s + Format(' Stop %1.30g (MaxDouble)',[ds]);
      AStringList.Add(s);
      AStringList.Add('Test succeeded');
      Result := True;
      Break;
    end;
    if (Abs(ds) > 0.5) and (Abs(ds) < 1e6) then
      s := s + Format(' Stop %1.30f',[ds])
    else
      s := s + Format(' Stop %1.30g',[ds]);
    AStringList.Add(s);

    if ds > 0 then
    begin
      if ds > MaxDouble / 2.0 then
        ds := ds - ((MaxDouble - ds) / 2.0)
      else
        ds := ds / 2.0;
    end
    else
    begin
      if ds < -MaxDouble / 2.0 then
        ds := ds - ((MaxDouble - Abs(ds)) / 2.0)
      else
        ds := ds * 2.0;
    end;
    Inc(cnt)
  until False;
end;

end.

